unit IdCoder3to4;

interface

uses
  Classes,
  IdCoder;

type
  TIdDecodeTable = array[1..127] of Byte;

  TIdEncoder3to4 = class(TIdEncoder)
  protected
    FCodingTable: string;
    FFillChar: Char;
  public
    function Encode(ASrcStream: TStream; const ABytes: integer = MaxInt): string; override;
    procedure EncodeUnit(const AIn1, AIn2, AIn3: Byte; var VOut: Cardinal);
  published
    property CodingTable: string read FCodingTable;
    property FillChar: Char read FFillChar write FFillChar;
  end;

  TIdEncoder3to4Class = class of TIdEncoder3to4;

  TIdDecoder4to3 = class(TIdDecoder)
  protected
    FDecodeTable: TIdDecodeTable;
    FFillChar: Char;
  public
    class procedure ConstructDecodeTable(const ACodingTable: string;
     var ADecodeArray: TIdDecodeTable);
    procedure DecodeToStream(AIn: string; ADest: TStream); override;
    procedure DecodeUnit(AIn: Cardinal; var VOut1, VOut2, VOut3: Byte);
  published
    property FillChar: Char read FFillChar write FFillChar;
  end;

implementation

uses
  IdException, IdGlobal, IdResourceStrings,
  SysUtils;

{ TIdDecoder4to3 }

class procedure TIdDecoder4to3.ConstructDecodeTable(const ACodingTable: string;
 var ADecodeArray: TIdDecodeTable);
var
  i: integer;
begin
  //TODO: See if we can find an efficient way, or maybe an option to see if the requested
  //decode char is valid, that is it returns a 255 from the DecodeTable, or at maybe
  //check its presence in the encode table.
  for i := Low(ADecodeArray) to High(ADecodeArray) do begin
    ADecodeArray[i] := 255;
  end;
  for i := 1 to Length(ACodingTable) do begin
    ADecodeArray[Ord(ACodingTable[i])] := i - 1;
  end;
end;

procedure TIdDecoder4to3.DecodeToStream(AIn: string; ADest: TStream);
var
  LOut: string;
  LOut1, LOut2, LOut3: Byte;
  LUnit: TIdCardinalBytes;
begin
  if (Length(AIn) mod 4) > 0 then begin
    raise EIdException.Create(RSUnevenSizeInDecodeStream);
  end;
  with TStringStream.Create(AIn) do try
    while Position < Size do begin
      ReadBuffer(LUnit.Whole, 4);
      DecodeUnit(LUnit.Whole, LOut1, LOut2, LOut3);
      // Must check Byte3 before for, as if Byte3 is FillChar, Byte 4 will be also be FillChar
      if Chr(LUnit.Byte3) = FillChar then begin
        LOut := Chr(LOut1);
      end else if Chr(LUnit.Byte4) = FillChar then begin
        LOut := Chr(LOut1) + Chr(LOut2);
      end else begin
        LOut := Chr(LOut1) + Chr(LOut2) + Chr(LOut3);
      end;
      ADest.WriteBuffer(LOut[1], Length(LOut));
    end;
  finally Free; end;
end;

procedure TIdDecoder4to3.DecodeUnit(AIn: Cardinal; var VOut1, VOut2, VOut3: Byte);
var
  LUnit: TIdCardinalBytes;
begin
  LUnit.Whole := AIn;
  LUnit.Whole := (FDecodeTable[LUnit.Byte1] shl 18) or (FDecodeTable[LUnit.Byte2] shl 12)
   or (FDecodeTable[LUnit.Byte3] shl 6) or FDecodeTable[LUnit.Byte4];
  VOut1 := LUnit.Byte3;
  VOut2 := LUnit.Byte2;
  VOut3 := LUnit.Byte1;
end;

{ TIdEncoder3to4 }

function TIdEncoder3to4.Encode(ASrcStream: TStream; const ABytes: integer = MaxInt): string;
var
  LIn1, LIn2, LIn3: Byte;
  LSize: integer;
  LStartPos: integer;
  LUnit: TIdCardinalBytes;
begin
  //TODO: Make this more efficient. Profile it to test, but maybe make single calls to ReadBuffer
  //then pull from memory
  if (ABytes <> MaxInt) and ((ABytes mod 3) > 0) then begin
    raise EIdException.Create(RSUnevenSizeInEncodeStream);
  end;
  Result := '';
  LStartPos := ASrcStream.Position;
  while (ASrcStream.Position < ASrcStream.Size) and (ASrcStream.Position - LStartPos < ABytes)
   do begin
    ASrcStream.ReadBuffer(LIn1, 1);
    if ASrcStream.Position < ASrcStream.Size then begin
      ASrcStream.ReadBuffer(LIn2, 1);
      if ASrcStream.Position < ASrcStream.Size then begin
        ASrcStream.ReadBuffer(LIn3, 1);
        LSize := 3;
      end else begin
        LIn3 := 0;
        LSize := 2;
      end;
    end else begin
      LIn2 := 0;
      LSize := 1;
    end;
    EncodeUnit(LIn1, LIn2, LIn3, LUnit.Whole);
    Result := Result + Chr(LUnit.Byte1) + Chr(LUnit.Byte2) + Chr(LUnit.Byte3) + Chr(LUnit.Byte4);
    if LSize < 3 then begin
      Result[Length(Result)] := FillChar;
      if LSize = 1 then begin
        Result[Length(Result) - 1] := FillChar;
      end;
    end;
  end;
end;

procedure TIdEncoder3to4.EncodeUnit(const AIn1, AIn2, AIn3: Byte; var VOut: Cardinal);
var
  LUnit: TIdCardinalBytes;
begin
  LUnit.Byte1 := Ord(FCodingTable[((AIn1 SHR 2) and 63) + 1]);
  LUnit.Byte2 := Ord(FCodingTable[(((AIn1 SHL 4) or (AIn2 SHR 4)) and 63) + 1]);
  LUnit.Byte3 := Ord(FCodingTable[(((AIn2 SHL 2) or (AIn3 SHR 6)) and 63) + 1]);
  LUnit.Byte4 := Ord(FCodingTable[(Ord(AIn3) and 63) + 1]);
  VOut := LUnit.Whole;
end;

end.
