unit IdHTTP;

{
  Implementation of the HTTP protcol as specified in RFC 2616, 2109, 2965.
  (See NOTE below for details of what is exactly implemented)

  Author: Hadi Hariri (hadi@urusoft.com)
  Copyright: (c) Chad Z. Hower and The Winshoes Working Group.

NOTE:
  Initially only GET and POST will be supported. As time goes on more will
  be added. For other developers, please add the date and what you have done
  below.

Initials: Hadi Hariri - HH

Details of implementation
-------------------------
2001-Jul-7 Doychin Bondzhev
 - new property AllowCookie
 - There is no more ExtraHeders property in Request/Response. Raw headers is used for that purpose.
2001-Jul-1 Doychin Bondzhev
 - SSL support is up again - Thanks to Gregor
2001-Jun-17 Doychin Bondzhev
 - New unit IdHTTPHeaderInfo.pas that contains the
   TIdHeaderInfo(TIdEntytiHeaderInfo, TIdRequestHeaderInfo and TIdResponseHeaderInfo)
 - Still in development and not verry well tested
   By default when there is no authorization object associated with HTTP compoenet and there is user name and password
   HTTP component creates and instance of TIdBasicAuthentication class. This behaivor is for both web server and proxy server
   authorizations
2001-Apr-17 Doychin Bondzhev
 - Added OnProxyAuthorization event. This event is called on 407 response from the HTTP Proxy.
 - Added 2 new properties in TIdHeaderInfo
    property AuthenticationScheme: TIdAuthenticationScheme - this property contains information for authentication scheme
      requested by the web server
    property ProxyAuthenticationScheme: TIdAuthenticationScheme - this property contains information for authentication scheme
      requested by the proxy server
 - Now the component authomaticly reconginizes the requested authorization scheme and it supports Basic like before and has been
   extend to support Digest authorization
2001-Mar-31 Doychin Bondzhev
 - If there is no CookieManager it does not support cookies.
2001-Feb-18 Doychin Bondzhev
 - Added OnAuthorization event. This event is called on 401 response from the HTTP server.
     This can be used to ask the user program to supply user name and password in order to acces
     the requested resource
2001-Feb-02 Doychin Bondzhev
 - Added Cookie support and relative paths on redirect
2000-Jul-25 Hadi Hariri
 - Overloaded POst and moved clearing to disconect.
2000-June-22 Hadi Hariri
  - Added Proxy support.
2000-June-10 Hadi Hariri
  - Added Chunk-Encoding support and HTTP version number. Some additional
    improvements.
2000-May-23 J. Peter Mugaas
  -added redirect capability and supporting properties.  Redirect is optional
   and is set with HandleRedirects.  Redirection is limited to RedirectMaximum
   to prevent stack overflow due to recursion and to prevent redirects between
   two places which would cause this to go on to infinity.
2000-May-22 J. Peter Mugaas
  -adjusted code for servers which returned LF instead of EOL
  -Headers are now retreived before an exception is raised.  This
   also facilitates server redirection where the server tells the client to
   get a document from another location.
2000-May-01 Hadi Hariri
  -Converted to Mercury
2000-May-01 Hadi Hariri
  -Added PostFromStream and some clean up
2000-Apr-10 Hadi Hariri
  -Re-done quite a few things and fixed GET bugs and finished POST method.
2000-Jan-13 MTL
  -Moved to the New Palette Scheme
2000-Jan-08 MTL
  -Cleaned up a few compiler hints during 7.038 build
1999-Dec-10 Hadi Hariri
  -Started.
}

interface

uses
  Classes,
  IdAssignedNumbers,
  IdHeaderList,
  IdHTTPHeaderInfo,
  IdIOHandlerSocket,
  IdSSLOpenSSL,
  IdTCPClient,
  IdURI,
  IdCookie,
  IdCookieManager,
  IdAuthentication;

type
  TIdHTTPMethod = (hmHead, hmGet, hmPost, hmOptions, hmTrace, hmPut, hmDelete);
  TIdHTTPProtocolVersion = (pv1_0, pv1_1);

  TIdHTTPOnRedirectEvent = procedure(Sender: TObject; var dest: String; var NumRedirect: Integer; var Handled: boolean; Var VMethod: TIdHTTPMethod) of object;
  TIdOnSelectAuthorization = procedure(Sender: TObject; Var AuthenticationClass: TIdAuthenticationClass; AuthInfo: TIdHeaderList) of object;
  TIdOnAuthorization = procedure(Sender: TObject; Var VUserName, VPassword: String; var Handled: boolean) of object;
  TIdProxyOnAuthorization = procedure(Sender: TObject; Var VUserName, VPassword: String; var Handled: boolean) of object;

const
  Id_TIdHTTP_ProtocolVersion = pv1_1;
  Id_TIdHTTP_RedirectMax = 15;
  Id_TIdHTTP_HandleRedirects = False;
  Id_TIdHTTP_MaxAuthRetries = 3;

type
  TIdHTTP = class(TIdTCPClient)
  protected
    FCookieManager: TIdCookieManager;
    FFreeOnDestroy: Boolean;
    FAllowCookies: Boolean; 

    FProtocolVersion: TIdHTTPProtocolVersion;
    FRedirectCount: Integer;
    FRedirectMax: Integer;
    FAtuhRetriesMax: Integer;
    FProxyAtuhRetriesMax: Integer;
    FRequest: TIdRequestHeaderInfo;
    FResponse: TIdResponseHeaderInfo;
    FResponseCode: Integer;
    FResponseText: string;
    FHandleRedirects : Boolean;
    FTunnelProxyHost: string;
    FTunnelProxyPort: Integer;
    FTunnelProxyProtocol: string;

    FURI: TIdURI;

    FCloseConnection: Boolean;

    {this is an internal counter for redirercts}
    //
    FOnRedirect: TIdHTTPOnRedirectEvent;
    FOnSelectAuthorization: TIdOnSelectAuthorization;
    FOnSelectProxyAuthorization: TIdOnSelectAuthorization;
    FOnAuthorization: TIdOnAuthorization;
    FOnProxyAuthorization: TIdProxyOnAuthorization;
    //

    function DoOnAuthorization: Boolean; virtual;
    function DoOnProxyAuthorization: Boolean; virtual;
    function DoOnRedirect(var Location: String; Var VMethod: TIdHTTPMethod; RedirectCount: integer): boolean; virtual;
    procedure DoProxyConnectMethod(ASender: TIdSSLIOHandlerSocket);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure ProcessCookies;
    procedure RetrieveHeaders;
    procedure SetCookieManager(ACookieManager: TIdCookieManager);
    procedure SetAllowCookies(AValue: Boolean);
    procedure SetCookies;
    function SetHostAndPort(AProto, AHost: string; APort: Integer): Boolean;
    procedure ReadResult(AResponseContent: TStream);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoRequest(const AMethod: TIdHTTPMethod; AURL: string;
      const ASource, AResponseContent: TStream); virtual;
    procedure Options(AURL: string); overload;
    procedure Get(AURL: string; const AResponseContent: TStream); overload;
    function Get(AURL: string): string; overload;
    procedure Trace(AURL: string; const AResponseContent: TStream); overload;
    function Trace(AURL: string): string; overload;
    procedure Head(AURL: string);

    function Post(AURL: string; const ASource: TStrings): String; overload;
    function Post(AURL: string; const ASource: TStream): String; overload;
    procedure Post(AURL:string; const ASource: TStrings; const AResponseContent:TStream);
      overload;
    {Post data provided by a stream, this is for submitting data to a server}
    procedure Post(AURL: string; const ASource, AResponseContent: TStream);
      overload;
    //
    function Put(AURL: string; const ASource: TStream): String; overload;
    procedure Put(AURL: string; const ASource, AResponseContent: TStream);
      overload;
    {This is the response code number such as 404 for File not Found}
    property ResponseCode: Integer read FResponseCode;
    {This is the text of the message such as "404 File Not Found here Sorry"}
    property ResponseText: string read FResponseText;
    property Response: TIdResponseHeaderInfo read FResponse;
    { This is the last processed URL }
    property URL: TIdURI read FURI;
  published
    property AllowCookies: Boolean read FAllowCookies write SetAllowCookies;
    {Do we handle redirect requests or simply raise an exception and let the
     developer deal with it}
    property HandleRedirects : Boolean read FHandleRedirects
      write FHandleRedirects default Id_TIdHTTP_HandleRedirects;
    {This is the maximum number of redirects we wish to handle, we limit this
     to prevent stack overflow due to recursion.  Recursion is safe ONLY if
     prevented for continuing to infinity}
    property ProtocolVersion: TIdHTTPProtocolVersion read FProtocolVersion write FProtocolVersion Default Id_TIdHTTP_ProtocolVersion;
    property RedirectMaximum: Integer read FRedirectMax write FRedirectMax default Id_TIdHTTP_RedirectMax;
    property AuthenticationRetriesMaximum: Integer read FAtuhRetriesMax write FAtuhRetriesMax default Id_TIdHTTP_MaxAuthRetries;
    property ProxyAuthMaxRetries: Integer read FProxyAtuhRetriesMax write FProxyAtuhRetriesMax default Id_TIdHTTP_MaxAuthRetries;
    property Request: TIdRequestHeaderInfo read FRequest write FRequest;
    // Fired when a rediretion is requested.
    property OnRedirect: TIdHTTPOnRedirectEvent read FOnRedirect write FOnRedirect;
    property OnSelectAuthorization: TIdOnSelectAuthorization read FOnSelectAuthorization write FOnSelectAuthorization;
    property OnSelectProxyAuthorization: TIdOnSelectAuthorization read FOnSelectProxyAuthorization write FOnSelectProxyAuthorization;
    property OnAuthorization: TIdOnAuthorization read FOnAuthorization write FOnAuthorization;
    property OnProxyAuthorization: TIdProxyOnAuthorization read FOnProxyAuthorization write FOnProxyAuthorization;
    property Host;
    property Port default IdPORT_HTTP;

    // Cookie stuff
    property CookieManager: TIdCookieManager read FCookieManager write SetCookieManager;
  end;

implementation

uses
  IdComponent, IdCoderMIME,
  IdTCPConnection,
  IdException,
  IdGlobal,
  IdResourceStrings,
  SysUtils;

const
  ProtocolVersionString: array[TIdHTTPProtocolVersion] of string = ('1.0', '1.1');

{ TIdHTTP }

constructor TIdHTTP.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Port := IdPORT_HTTP;

  AllowCookies := true;
  FFreeOnDestroy := false;

  FRedirectMax := Id_TIdHTTP_RedirectMax;
  FHandleRedirects := Id_TIdHTTP_HandleRedirects;
  FAtuhRetriesMax := Id_TIdHTTP_MaxAuthRetries;
  FProxyAtuhRetriesMax := Id_TIdHTTP_MaxAuthRetries;

  // Create the headers
  FRequest := TIdRequestHeaderInfo.Create{(self)};
  FResponse := TIdResponseHeaderInfo.Create{(Self)};
  //
  FProtocolVersion := Id_TIdHTTP_ProtocolVersion;

  FURI := TIdURI.Create('');
end;

destructor TIdHTTP.Destroy;
begin
  FreeAndNil(FRequest);
  FreeAndNil(FResponse);
  FreeAndNil(FURI);

  if FFreeOnDestroy then
  begin
    FreeAndNil(FCookieManager);
  end;

  inherited Destroy;
end;

procedure TIdHTTP.Options(AURL: string);
begin
  DoRequest(hmOptions, AURL, nil, nil);
end;

procedure TIdHTTP.Get(AURL: string; const AResponseContent: TStream);
begin
  DoRequest(hmGet, AURL, nil, AResponseContent);
end;

procedure TIdHTTP.Trace(AURL: string; const AResponseContent: TStream);
begin
  DoRequest(hmTrace, AURL, nil, AResponseContent);
end;

procedure TIdHTTP.Head(AURL: string);
begin
  DoRequest(hmHead, AURL, nil, nil);
end;

procedure TIdHTTP.Post(AURL: string; const ASource, AResponseContent: TStream);
var
  OldProtocol: TIdHTTPProtocolVersion;
begin
  // PLEASE READ CAREFULLY

  // Currently when issuing a POST, IdHTTP will automatically set the protocol
  // to version 1.0 independently of the value it had initially. This is because
  // there are some servers that don't respect the RFC to the full extent. In
  // particular, they don't respect sending/not sending the Expect: 100-Continue
  // header. Until we find an optimum solution that does NOT break the RFC, we
  // will restrict POSTS to version 1.0.
  if Connected then
  begin
    Disconnect;
  end;
  OldProtocol := FProtocolVersion;
  FProtocolVersion := pv1_0;
  DoRequest(hmPost, AURL, ASource, AResponseContent);
  FProtocolVersion := OldProtocol;
end;

procedure TIdHTTP.Post(AURL: string; const ASource: TStrings; const AResponseContent: TStream);
Var
  LParams: TStringStream;
  S: String;
begin
  S := StringReplace(ASource.Text, EOL, '&', [rfReplaceall]);
  LParams := TStringStream.Create(S);
  try
    Post(AURL, LParams, AResponseContent);
  finally
    LParams.Free;
  end;
end;

function TIdHTTP.Post(AURL: string; const ASource: TStrings): String;
Var
  LResponse: TStringStream;
  LParams: TStringStream;
  S: String;
begin
  LResponse := TStringStream.Create('');
  S := StringReplace(ASource.Text, EOL, '&', [rfReplaceall]);
  LParams := TStringStream.Create(S);
  try
    Post(AURL, LParams, LResponse);
  finally
    result := LResponse.DataString;
    LResponse.Free;
    LParams.Free;
  end;
end;

function TIdHTTP.Post(AURL: string; const ASource: TStream): String;
Var
  LResponse: TStringStream;
begin
  LResponse := TStringStream.Create('');
  try
    Post(AURL, ASource, LResponse);
  finally
    result := LResponse.DataString;
    LResponse.Free;
  end;
end;

procedure TIdHTTP.Put(AURL: string; const ASource, AResponseContent: TStream);
begin
  DoRequest(hmPut, AURL, ASource, AResponseContent);
end;

function TIdHTTP.Put(AURL: string; const ASource: TStream): String;
Var
  LResponse: TStringStream;
begin
  LResponse := TStringStream.Create('');
  try
     Put(AURL, ASource, LResponse);
  finally
    result := LResponse.DataString;
    LResponse.Free;
  end;
end;

function TIdHTTP.Get(AURL: string): string;
var
  Stream: TStringStream;
begin
  Stream := TStringStream.Create(''); try
    Get(AURL, Stream);
  finally
    result := Stream.DataString;
    Stream.Free;
  end;
end;

function TIdHTTP.Trace(AURL: string): string;
var
  Stream: TStringStream;
begin
  Stream := TStringStream.Create(''); try
    Trace(AURL, Stream);
    result := Stream.DataString;
  finally Stream.Free; end;
end;

procedure TIdHTTP.RetrieveHeaders;
var
  s:string;
begin
  // Set the response headers
  // Clear headers
  // Don't use Capture.

  FResponse.RawHeaders.Clear;
  s := ReadLn;
  while Length(s) > 0 do
  begin
    FResponse.RawHeaders.Add(S);
    try
      s := ReadLn;
    except
      on E: EIdConnClosedGracefully do
      begin
        Disconnect;
        break;
      end;
    end;
  end;
  FResponse.ProcessHeaders;

  ProcessCookies;
end;

function TIdHTTP.DoOnRedirect(var Location: String; Var VMethod: TIdHTTPMethod; RedirectCount: integer): boolean;
begin
  result := HandleRedirects;
  if assigned(FOnRedirect) then
  begin
    FOnRedirect(self, Location, RedirectCount, result, VMethod);
  end;
end;

procedure TIdHTTP.SetCookies;
Var
  S: String;
begin
  if Assigned(FCookieManager) then
  begin
    // Send secure cookies only if we have Secured connection
    S := FCookieManager.GenerateCookieList(URL, (IOHandler is TIdSSLIOHandlerSocket));
    if Length(S) > 0 then
    begin
      Request.RawHeaders.Values['Cookie'] := S;
    end;
  end;
end;

// This function sets the Host and Port and returns a boolean depending on
// whether a PROXY is being used or not.
function TIdHTTP.SetHostAndPort(AProto, AHost: string; APort: Integer): Boolean;
begin
  // First check to see if a Proxy has been specified.
  if Length(Request.ProxyParams.ProxyServer) > 0 then
  begin
    if ((not AnsiSameText(Host, Request.ProxyParams.ProxyServer)) or
       (Port <> Request.ProxyParams.ProxyPort)) and (Connected) then
    begin
      Disconnect;
    end;

    Host := Request.ProxyParams.ProxyServer;
    Port := Request.ProxyParams.ProxyPort;

    if Length(AHost) > 0 then
    begin
      FURI.Host := AHost;
    end;

    if Length(AProto) > 0 then
    begin
      FURI.Protocol := AProto;
    end;

    if APort <> -1 then
    begin
      FURI.Port := IntToStr(APort);
    end;

    if AnsiSameText(AProto, 'HTTPS') then
    begin
      FTunnelProxyHost := AHost;

      if APort = -1 then
      begin
        FTunnelProxyPort := IdPORT_SSL
      end
      else begin
        FTunnelProxyPort := APort;
      end;

      FURI.Port := IntToStr(FTunnelProxyPort);
      FTunnelProxyProtocol := ProtocolVersionString[ProtocolVersion];
      Result := False;

      if Assigned(IOHandler) then
      begin
        if (IOHandler is TIdSSLIOHandlerSocket) then
        begin
          // Send the CONNECT command to proxy server, to pass-through the communication
          TIdSSLIOHandlerSocket(IOHandler).OnBeforeConnect := DoProxyConnectMethod;
        end
        else begin
          raise EIdInterceptPropInvalid.Create(RSIOHandlerPropInvalid);
        end;
      end;
    end
    else begin
      FURI.Port := IntToStr(IdPORT_HTTP);
      Result := True;
    end;
  end
  else begin
    Result := False;
    if AnsiSameText(AProto, 'HTTPS') then
    begin
      if Assigned(IOHandler) then
      begin
        if IOHandler is TIdSSLIOHandlerSocket then
        begin
          TIdSSLIOHandlerSocket(IOHandler).OnBeforeConnect := nil;

          if (Length(AHost) > 0) then
          begin
            if ((not AnsiSameText(Host, AHost)) or (Port<>APort)) and (Connected) then
            begin
              Disconnect;
            end;
            Host := AHost;
            FURI.Host := AHost;
            if Length(AProto) > 0 then FURI.Protocol := AProto;
          end
          else
            FURI.Host := Host;
          // Check to see if a PORT has been specified
          if APort = -1 then
            Port := IdPORT_SSL
          else
            Port := APort;
        end
        else begin
          // IOHandler is not TIdIOSSLHandlerSocket
          raise EIdInterceptPropInvalid.Create(RSIOHandlerPropInvalid);
        end;
      end
      else begin
        // IOHandler is nil
        raise EIdInterceptPropInvalid.Create(RSIOHandlerPropInvalid);
      end;
    end
    else
    begin
      if Assigned(IOHandler) then
      begin
        if (IOHandler is TIdSSLIOHandlerSocket) then
        begin
          (IOHandler as TIdSSLIOHandlerSocket).PassThrough := true;
        end;
      end;

      if (Length(AHost) > 0) then
      begin
        if APort = -1 then
          APort := IdPORT_HTTP;

        if ((not AnsiSameText(Host, AHost)) or (Port <> APort)) and (Connected) then
        begin
          Disconnect;
        end;
        Host := AHost;
        FURI.Host := AHost;
        if Length(AProto) > 0 then FURI.Protocol := AProto;
      end
      else
        FURI.Host := Host;
      if APort = -1 then
      begin
        Port := IdPORT_HTTP;
      end
      else begin
        Port := APort;
      end;
    end;

    FURI.Port := IntToStr(Port);
  end;
end;

procedure TIdHTTP.ReadResult(AResponseContent: TStream);
var
  Size: Integer;

  function ChunkSize: integer;
  var
    j: Integer;
    s: string;
  begin
    s := ReadLn;
    j := AnsiPos(' ', s);
    if j > 0 then
    begin
      s := Copy(s, 1, j - 1);
    end;
    Result := StrToIntDef('$' + s, 0);
  end;

begin
  if Assigned(AResponseContent) then // Only for Get and Post
  begin
    if Response.ContentLength <> 0 then // If chunked then this is also 0
    begin
      ReadStream(AResponseContent, Response.ContentLength);
    end
    else
    begin
      if AnsiPos('chunked', Response.RawHeaders.Values['Transfer-Encoding']) > 0 then {do not localize}
      begin // Chunked
        DoStatus(hsStatusText,  [RSHTTPChunkStarted]);
        Size := ChunkSize;
        while Size > 0 do
        begin
          ReadStream(AResponseContent, Size);
          ReadLn; // blank line
          Size := ChunkSize;
        end;
        ReadLn; // blank line
      end
      else begin
        if not Response.HasContentLength then
          ReadStream(AResponseContent, -1, True);
      end;
    end;
  end;
end;

procedure TIdHTTP.DoRequest(const AMethod: TIdHTTPMethod; AURL: string;
 const ASource, AResponseContent: TStream);
var
  LUserName: string;
  LPassword: string;
  LLocation: string;
  LDoc: string;
  LHost: string;
  LPath: string;
  LProto: string;
  LPort: string;
  LBookmark : string;
  LURL: string;
  LResponseVersion: string;
  ResponseDigit: Integer;
  LResponseLocation: Integer;
  i: Integer;
  LMethod: TIdHTTPMethod;

  procedure PrepaereRequest;
  begin
    LURL := AURL;

    TIdURI.ParseURI({URLDecode(}AURL{)}, LProto, LHost, LPath, LDoc, LPort, LBookmark, LUserName, LPassword);

    if Length(LUserName) > 0 then
    begin
      Request.Username := LUserName;
      Request.Password := LPassword;
    end;

    FURI.Username := Request.Username;
    FURI.Password := Request.Password;

    FURI.Path := ProcessPath(FURI.Path, LPath);
    FURI.Document := LDoc;

    // The URL part is not URL encoded at this place

    AURL := {URLEncode(}LPath + LDoc{)};

    if AMethod = hmOptions then
    begin
      if AnsiSameText(LDoc, '*') then
      begin
        AURL := LDoc;
      end;
    end;

    if (Length(LHost) = 0) and (Length(Host) > 0) then
    begin
      LHost := Host;
    end;

    if StrToIntDef(LPort, -1) = -1 then
    begin
      if Length(LProto) = 0 then
      begin
        LPort := IntToStr(Port);
      end;
    end;

    if SetHostAndPort(LProto, LHost, StrToIntDef(LPort, -1)) then
    begin
      if Length(LHost) > 0 then
      begin
        AURL := LHost + AURL
      end
      else begin
        AURL := FURI.Host + AURL;
      end;

      if Length(LProto) > 0 then
      begin
        AURL := LProto + '://' + AURL
      end
      else begin
        AURL := FURI.Protocol + '://' +  AURL;
      end;
    end;

    // Check for valid HTTP request methods
    if AMethod in [hmTrace, hmPut, hmOptions, hmDelete] then
    begin
      if ProtocolVersion <> pv1_1 then
      begin
        raise EIdException.Create('This request method is supported in HTTP 1.1');
      end;
    end;

    if AMethod in [hmPost, hmPut] then
    begin
      Request.ContentLength := TStream(ASource).Size;
    end
    else Request.ContentLength := 0;

    Request.Host := FURI.Host;
  end;

begin
  Inc(FRedirectCount);

  if Assigned(AResponseContent) then
  begin
    LResponseLocation := AResponseContent.Position;
  end
  else
    LResponseLocation := 0;  // Just to avoid the waringing message

  PrepaereRequest;

  if not Connected then try
    Connect;
  except
    on E: EIdSSLProtocolReplyError do
    begin
      Disconnect;
      raise;
    end;
  end;

  Request.SetHeaders;
  SetCookies;

  // This is a wrokaround for some HTTP servers wich does not implement properly the HTTP protocol
  OpenWriteBuffer;
  case AMethod of
    hmHead: WriteLn('HEAD ' + AURL + ' HTTP/' + ProtocolVersionString[ProtocolVersion]); {do not localize}
    hmGet: WriteLn('GET ' + AURL  + ' HTTP/' + ProtocolVersionString[ProtocolVersion]); {do not localize}
    hmPost: WriteLn('POST ' + AURL + ' HTTP/'+ ProtocolVersionString[ProtocolVersion]); {do not localize}
    // HTTP 1.1 only
    hmOptions: WriteLn('OPTIONS ' + AURL + ' HTTP/'+ ProtocolVersionString[ProtocolVersion]); {do not localize}
    hmTrace: WriteLn('TRACE ' + AURL + ' HTTP/'+ ProtocolVersionString[ProtocolVersion]); {do not localize}
    hmPut: WriteLn('PUT ' + AURL + ' HTTP/'+ ProtocolVersionString[ProtocolVersion]); {do not localize}
  end;
  // write the headers
  for i := 0 to Request.RawHeaders.Count - 1 do
    if Length(Request.RawHeaders.Strings[i]) > 0 then 
      WriteLn(Request.RawHeaders.Strings[i]);
  WriteLn('');
  CloseWriteBuffer;

  if (AMethod in [hmPost, hmPut]) then
  begin
    WriteStream(TStream(ASource), True, false);
  end;

  // Workaround for servers wich respond with 100 Continue on GET and HEAD
  // This workaround is just for temporary use until we have final HTTP 1.1
  // realisation
  repeat
    FResponseText := ReadLn;
    LResponseVersion := Copy(FResponseText, 6, 3);
    RetrieveHeaders;

    // CloseConnection will be true in one of the following cases:
    // 1. Response.Connection or Response.ProxyConnection were Close
    // 2. We have 1.0 response without Connection: Keep-Alive
    // 3. when it is closed by the server

    FCloseConnection := (AnsiSameText(Trim(Response.Connection), 'CLOSE') or
      AnsiSameText(Trim(Response.ProxyConnection), 'CLOSE')) and Connected;

    if not FCloseConnection then
      with Response do
      begin
        case ProtocolVersion of
          pv1_1:
            begin
              FCloseConnection := 
                ((not AnsiSameText(Trim(Connection), 'KEEP-ALIVE')) and
                (not AnsiSameText(Trim(ProxyConnection), 'KEEP-ALIVE'))) and
                (not (AnsiSameText(LResponseVersion, '1.1')));
            end;
          pv1_0:
            begin
              FCloseConnection := (not AnsiSameText(Trim(Connection), 'KEEP-ALIVE')) or
                not ((AnsiSameText(Trim(Connection), '') and AnsiSameText(Trim(ProxyConnection), 'KEEP-ALIVE')));
            end;
        end;
      end;

    Fetch(FResponseText);
    FResponseCode := StrToInt(Fetch(FResponseText,' ', False));
  until FResponseCode <> 100;

  ResponseDigit := ResponseCode div 100;

  try
    // Handle Redirects
    if ((ResponseDigit = 3) and (ResponseCode <> 304)) or (Length(Response.Location) > 0) then
    begin
      // First read the result
      if Connected {and (Response.ContentLength > 0)} then
      begin
        // This is a workaround for buggy HTTP 1.1 servers which
        // does not return any body with 302 response code
        LResponseLocation := ReadTimeout;
        ReadTimeout := 2000;
        try
          ReadResult(AResponseContent);
        except end;
        ReadTimeout := LResponseLocation;
      end;

      LLocation := URLDecode(Response.Location);

      if (FHandleRedirects) and (FRedirectCount < FRedirectMax) then
      begin
        if Assigned(AResponseContent) then
        begin
          AResponseContent.Position := LResponseLocation;
        end;

        LMethod := AMethod;
        if DoOnRedirect(LLocation, LMethod, FRedirectCount) then
        begin
          if FCloseConnection then
          begin
            Disconnect;
          end;

          DoRequest(LMethod, LLocation, ASource, AResponseContent);
        end;
      end
      else // Just fire the event
      begin
        LMethod := AMethod;
        if not DoOnRedirect(LLocation, LMethod, FRedirectCount) then // If not Handled
          raise EIdProtocolReplyError.CreateError(ResponseCode, ResponseText)
        else Response.Location := LLocation;
      end;
    end
    else
    begin
      if (ResponseCode <> 204) and Connected then
      begin
        ReadResult(AResponseContent);
      end;

      // GREGOR Workaround
      // if we get an error we disconnect if we use SSLIOHandler
      if Assigned(IOHandler) then
      begin
        FCloseConnection := Connected and (IOHandler is TIdSSLIOHandlerSocket) or FCloseConnection;
      end;

      if ResponseDigit <> 2 then
      begin
        if FCloseConnection then
        begin
          Disconnect;
        end;

        case ResponseCode of
          401:
            begin // HTTP Server authorization requered
              if DoOnAuthorization then
              begin
                if Assigned(AResponseContent) then
                begin
                  AResponseContent.Position := LResponseLocation;
                end;
                DoRequest(AMethod, LURL, ASource, AResponseContent);
              end;
            end;
          407:
            begin // Proxy Server authorization requered
              if DoOnProxyAuthorization then
              begin
                if Assigned(AResponseContent) then
                begin
                  AResponseContent.Position := LResponseLocation;
                end;
                DoRequest(AMethod, LURL, ASource, AResponseContent);
              end;
            end;
          else begin
            raise EIdProtocolReplyError.CreateError(ResponseCode, ResponseText);
          end;
        end;
      end;
    end;
  finally
    if FCloseConnection then
    begin
      Disconnect;
    end;

    if not Connected and Assigned(Request.Authentication) then
    begin
      Request.Authentication.Free;
      Request.Authentication := nil;
    end;

    FRedirectCount := 0;
  end;
end;

procedure TIdHTTP.DoProxyConnectMethod(ASender: TIdSSLIOHandlerSocket);
var
  sPort: String;
  S, LResponseText: String;
  LResponseCode: Integer;
begin
  // Turn off SSL for now.
  TIdSSLIOHandlerSocket(IOHandler).PassThrough := True;

  // Send the CONNECT command, if it is 200 OK then go on else disconnect
  sPort := IntToStr(FTunnelProxyPort);

  repeat
    WriteLn('CONNECT ' + FTunnelProxyHost + ':' + sPort + ' HTTP/' + FTunnelProxyProtocol);
    if Assigned(Request.ProxyParams.Authentication) then
    begin
      S := Request.ProxyParams.Authentication.Authentication;
      if Length(S) > 0 then
      begin
        WriteLn('Proxy-Authorization: ' + S);
      end;
    end;
    WriteLn;

    LResponseText := ReadLn;
    Fetch(LResponseText);
    LResponseCode := StrToInt(Fetch(LResponseText,' ', False));

    while Length(ReadLn) > 0 do ;

    if not ((ResponseCode = 407) and DoOnProxyAuthorization) then
    begin
      if LResponseCode <> 200 then
      begin
        raise EIdSSLProtocolReplyError.CreateError(LResponseCode, LResponseText);
      end
      else begin
        break;
      end;
    end;
  until false;

  // Turn on SSL
  TIdSSLIOHandlerSocket(IOHandler).PassThrough := False;
end;

procedure TIdHTTP.SetAllowCookies(AValue: Boolean);
begin
  FAllowCookies := AValue;
end;

procedure TIdHTTP.ProcessCookies;
Var
  // S: String;
  Cookies, Cookies2: TStringList;
  i: Integer;
begin
  if not Assigned(FCookieManager) and AllowCookies then
  begin
    CookieManager := TIdCookieManager.Create(Owner);
    FFreeOnDestroy := true;
  end;

  if Assigned(FCookieManager) then
  begin
    Request.ContentType := '';

    Cookies := TStringList.Create;
    Cookies2 := TStringList.Create;

    Response.RawHeaders.Extract('Set-cookie', Cookies);
    Response.RawHeaders.Extract('Set-cookie2', Cookies2);

    for i := 0 to Cookies.Count - 1 do
      CookieManager.AddCookie(Cookies[i], FURI.Host);

    for i := 0 to Cookies2.Count - 1 do
      CookieManager.AddCookie2(Cookies2[i], FURI.Host);

    {with TIdHeaderList.Create do
    begin
      Assign(Response.RawHeaders);
      while true do
      begin
        S := Values['Set-cookie'];
        if Length(S) > 0 then
        begin
          Delete(IndexOfName('Set-cookie'));
          CookieManager.AddCookie(S, FURI.Host);
        end
        else begin
          // Check for cookies as described in RFC 2965
          S := Values['Set-cookie2'];
          if Length(S) > 0 then
          begin
            Delete(IndexOfName('Set-cookie2'));
            CookieManager.AddCookie2(S, FURI.Host);
          end
          else break;
        end;
      end;
      Free;
    end;}
  end;
end;

procedure TIdHTTP.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (AComponent = FCookieManager) then
    begin
      FCookieManager := nil;
    end;
  end;
end;

procedure TIdHttp.SetCookieManager(ACookieManager: TIdCookieManager);
begin
  if Assigned(FCookieManager) then
  begin
    if FFreeOnDestroy then FCookieManager.Free;
  end;

  FCookieManager := ACookieManager;
  FFreeOnDestroy := false;

  if Assigned(FCookieManager) then
  begin
    FCookieManager.FreeNotification(Self);
  end;
end;

function TIdHttp.DoOnAuthorization: Boolean;
Var
  LUserName: String;
  LPassword: String;
  i: Integer;
  S: String;
  Auth: TIdAuthenticationClass;
begin
  if not Assigned(Request.Authentication) then
  begin
    // Find wich Authentication method is supported from us.
    for i := 0 to Response.WWWAuthenticate.Count - 1 do
    begin
      S := Response.WWWAuthenticate[i];
      try
        Auth := FindAuthClass(Fetch(S));
        break;
      except
      end;
    end;

    if i = Response.WWWAuthenticate.Count then
    begin
      result := false;
      exit;
    end;

    if Assigned(FOnSelectAuthorization) then
    begin
      OnSelectAuthorization(self, Auth, Response.WWWAuthenticate);
    end;

    Request.Authentication := Auth.Create;
  end;

  result := (Request.Authentication.Retries < FAtuhRetriesMax) and Assigned(FOnAuthorization);

  if Result then
  begin
    Request.Authentication.Username := Request.Username;
    Request.Authentication.Password := Request.Password;

    Request.Authentication.SetAuthParams(Response.WWWAuthenticate);

    result := false;

    repeat
      case Request.Authentication.Next of
        wnAskTheProgram:
          begin  // Ask the user porgram to supply us with authorization information
            if Assigned(FOnAuthorization) then
            begin
              LUserName := Request.Username;
              LPassword := Request.Password;

              OnAuthorization(self, LUserName, LPassword, result);

              if result then
              begin
                Request.Authentication.Username := LUserName;
                Request.Authentication.Password := LPassword;
                Request.Username := LUserName;
                Request.Password := LPassword;
              end
              else begin
                break;
              end;
            end;
          end;
        wnDoRequest:
          begin
            result := true;
            Request.Connection := 'Keep-Alive';
            break;
          end;
        wnFail:
          begin
            result := False;
            Request.Connection := 'Close';
            Break;
          end;
      end;
    until false;
  end;
end;

function TIdHttp.DoOnProxyAuthorization: Boolean;
Var
  LUserName: String;
  LPassword: String;
  i: Integer;
  S: String;
  Auth: TIdAuthenticationClass;
begin
  if not Assigned(Request.ProxyParams.Authentication) then
  begin
    // Find wich Authentication method is supported from us.
    for i := 0 to Response.ProxyAuthenticate.Count - 1 do
    begin
      S := Response.ProxyAuthenticate[i];
      try
        Auth := FindAuthClass(Fetch(S));
        break;
      except
      end;
    end;

    if i = Response.ProxyAuthenticate.Count then
    begin
      result := false;
      exit;
    end;

    if Assigned(FOnSelectProxyAuthorization) then
    begin
      OnSelectProxyAuthorization(self, Auth, Response.ProxyAuthenticate);
    end;

    Request.ProxyParams.Authentication := Auth.Create;
  end;

  result := (Request.ProxyParams.Authentication.Retries < FProxyAtuhRetriesMax) and Assigned(OnProxyAuthorization);

  if Result then
  begin
    Request.ProxyParams.Authentication.Username := Request.ProxyParams.ProxyUsername;
    Request.ProxyParams.Authentication.Password := Request.ProxyParams.ProxyPassword;

    Request.ProxyParams.Authentication.SetAuthParams(Response.ProxyAuthenticate);

    result := false;

    repeat
      case Request.ProxyParams.Authentication.Next of
        wnAskTheProgram: // Ask the user porgram to supply us with authorization information
          begin
            if Assigned(OnProxyAuthorization) then
            begin
              LUserName := Request.ProxyParams.ProxyUsername;
              LPassword := Request.ProxyParams.ProxyPassword;

              OnProxyAuthorization(self, LUserName, LPassword, result);

              if result then
              begin
                Request.ProxyParams.Authentication.Username := LUserName;
                Request.ProxyParams.Authentication.Password := LPassword;
              end
              else begin
                break;
              end;
            end;
          end;
        wnDoRequest:
          begin
            result := true;
            break;
          end;
        wnFail:
          begin
            result := False;
            Request.Connection := 'Close';
            Break;
          end;
      end;
    until false;
  end;
end;

end.
