unit IdURI;

{Details of implementation
-------------------------

2001-Feb-18 Doychin Bondzhev
 - Added UserName and Password to support URI's like
    http://username:password@hostname:port/path/document#bookmark
}

interface

type
  TIdURI = class
  protected
    FDocument: string;
    FProtocol: string;
    FURI: String;
    FPort: string;
    Fpath: string;
    FHost: string;
    FBookmark: string;
    FUserName: string;
    FPassword: string;
    FParams: string;
    //
    procedure SetURI(const Value: String);
    function GetURI: String;
  public
    constructor Create(const AURI: string = ''); virtual;
    class procedure NormalizePath(var APath: string);
    class procedure ParseURI(AURI: string; var VProtocol, VHost, VPath, VDocument, VPort, VBookmark,
      VUserName, VPassword: string); overload;
    class procedure ParseURI(AURI: string; var VProtocol, VHost, VPath, VDocument, VPort, VBookmark
      : string); overload;
    //
    property Bookmark : string read FBookmark write FBookMark;
    property Document: string read FDocument write FDocument;
    property Host: string read FHost write FHost;
    property Password: string read FPassword write FPassword;
    property Path: string read FPath write FPath;
    property Port: string read FPort write FPort;
    property Protocol: string read FProtocol write FProtocol;
    property URI: string read GetURI write SetURI;
    property Username: string read FUserName write FUserName;
  end;

implementation

uses
  IdGlobal, IdResourceStrings,
  SysUtils;

constructor TIdURI.Create(const AURI: string = '');
begin
  if length(AURI) > 0 then begin
    URI := AURI;
  end;
end;

class procedure TIdURI.NormalizePath(var APath: string);
var
  i: Integer;
begin
  // Normalize the directory delimiters to follow the UNIX syntax
  i := 1;
  while i <= Length(APath) do begin
    if APath[i] in LeadBytes then begin
      inc(i, 2)
    end else if APath[i] = '\' then begin
      APath[i] := '/';
      inc(i, 1);
    end else begin
      inc(i, 1);
    end;
  end;
end;

class procedure TIdURI.ParseURI(AURI: string; var VProtocol, VHost, VPath, VDocument, VPort, VBookmark
  : string);
Var
  LUsername: String;
  LPassword: String;
begin
  TIdURI.ParseURI(AURI, VProtocol, VHost, VPath, VDocument, VPort, VBookmark, LUserName, LPassword);
end;

// Used internally - don't include in the documentation
{const
  UnsafeChars = ['*', '#', '%', '<', '>', '+', ' '];  {do not localize}

// Used internally - don't include in the documentation
{function FixUnsafeChars(const ASrc: string): string;
var
  i: Integer;
begin
  Result := '';
  for i := 1 to Length(ASrc) do begin
    if (ASrc[i] in UnsafeChars) or (ASrc[i] >= #$80) then begin
      Result := Result + '%' + IntToHex(Ord(ASrc[i]), 2);  {do not localize}
{    end else begin
      Result := Result + ASrc[i];
    end;
  end;
end;}

class procedure TIdURI.ParseURI(AURI: string; var VProtocol, VHost, VPath, VDocument, VPort
 , VBookmark, VUserName, VPassword : string);
var
  LBuffer: string;
  LTokenPos, LPramsPos: Integer;
  LParams: String;
begin
  VHost := '';
  VProtocol := '';
  VPath := '';
  VDocument := '';
  NormalizePath(AURI);
  if IndyPos('://', AURI) > 0 then begin
    // absolute URI
    // What to do when data don't match configuration ??
    // Get the protocol
    LTokenPos := IndyPos('://', AURI);
    VProtocol := Copy(AURI, 1, LTokenPos  - 1);
    Delete(AURI, 1, LTokenPos + 2);
    // Get the user name, password, host and the port number
    LBuffer := Fetch(AURI, '/', True);
    // Get username and password
    LTokenPos := IndyPos('@', LBuffer);
    VPassword := Copy(LBuffer, 1, LTokenPos  - 1);
    if LTokenPos > 0 then
      Delete(LBuffer, 1, LTokenPos + 2);
    VUserName := Fetch(VPassword, ':', True);
    // Ignore cases where there is only password (http://:password@host/pat/doc)
    if Length(VUserName) = 0 then VPassword := '';
    // Get the host and the port number
    VHost := Fetch(LBuffer, ':', True);
    VPort := LBuffer;
    // Get the path
    LPramsPos := IndyPos('?', AURI);
    if LPramsPos > 0 then begin // The case when there is parameters after the document name '?'
      LTokenPos := RPos('/', AURI, LPramsPos);
    end
    else begin
      LPramsPos := IndyPos('=', AURI);
      if LPramsPos > 0 then begin // The case when there is parameters after the document name '='
        LTokenPos := RPos('/', AURI, LPramsPos);
      end
      else begin
        LTokenPos := RPos('/', AURI, -1);
      end;
    end;

    VPath := '/' + Copy(AURI, 1, LTokenPos);
    // Get the document
    if LPramsPos > 0 then begin
      VDocument := Copy(AURI, 1, LPramsPos - 1);
      Delete(AURI, 1, LPramsPos - 1);
      LParams := AURI;
    end
    else
      VDocument := AURI;
    Delete(VDocument, 1, LTokenPos);
  end else begin
    // received an absolute path, not an URI
    LPramsPos := IndyPos('?', AURI);
    if LPramsPos > 0 then begin // The case when there is parameters after the document name '?'
      LTokenPos := RPos('/', AURI, LPramsPos);
    end else begin
      LPramsPos := IndyPos('=', AURI);
      if LPramsPos > 0 then begin // The case when there is parameters after the document name '='
        LTokenPos := RPos('/', AURI, LPramsPos);
      end else begin
        LTokenPos := RPos('/', AURI, -1);
      end;
    end;

    VPath := Copy(AURI, 1, LTokenPos);
    // Get the document
    if LPramsPos > 0 then begin
      VDocument := Copy(AURI, 1, LPramsPos - 1);
      Delete(AURI, 1, LPramsPos - 1);
      LParams := AURI;
    end else begin
      VDocument := AURI;
    end;
    Delete(VDocument, 1, LTokenPos);
  end;

  // Parse the # bookmark from the document
  VBookmark := LParams;
  LParams := Fetch(VBookmark, '#');

  VPath := {FixUnsafeChars(}VPath{)};
  VDocument := {FixUnsafeChars(}VDocument{)} + {URLEncode(}LParams{)};
end;

procedure TIdURI.SetURI(const Value: String);
begin
  FURI := Value;
  NormalizePath(FURI);
  ParseURI(FURI, FProtocol, FHost, FPath, FDocument, FPort, FBookmark{, FUserName, FPassword});
end;

function TIdURI.GetURI: String;
begin
  FURI := FProtocol + '://';

  if Length(FUserName) > 0 then begin
    FURI := FURI + FUserName;

    if Length(FPassword) > 0 then begin
      FURI := FURI + ':' + FPassword;
    end;

    FURI := FURI + '@';
  end;

  FURI := FURI + FHost;
  if Length(FPort) > 0 then begin
    FURI := FURI + ':' + FPort;
  end;
  FURI := FURI + FPath + FDocument;
  if Length(FBookmark) > 0 then begin
    FURI := FURI + '#' + FBookmark;
  end;
  result := FURI;
end;

end.
